

set -ex



pip check
python -m jupyterhub --help-all
jupyterhub -h
configurable-http-proxy -h
test -f $PREFIX/share/jupyterhub/templates/page.html
test -f $PREFIX/share/jupyterhub/static/components/jquery/dist/jquery.min.js
test -f $PREFIX/share/jupyterhub/static/css/style.min.css
test -f $PREFIX/share/jupyterhub/static/js/home.js
exit 0
