



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -m jupyterhub --help-all
IF %ERRORLEVEL% NEQ 0 exit /B 1
jupyterhub -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
configurable-http-proxy -h
IF %ERRORLEVEL% NEQ 0 exit /B 1
test -f $PREFIX/share/jupyterhub/templates/page.html
IF %ERRORLEVEL% NEQ 0 exit /B 1
test -f $PREFIX/share/jupyterhub/static/components/jquery/dist/jquery.min.js
IF %ERRORLEVEL% NEQ 0 exit /B 1
test -f $PREFIX/share/jupyterhub/static/css/style.min.css
IF %ERRORLEVEL% NEQ 0 exit /B 1
test -f $PREFIX/share/jupyterhub/static/js/home.js
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
